"""In this example, we measure the jitter/resolution of the Time Tagger channels with the built-in test signals
and compare the results to the specifications of the Time Tagger, which is connected.
"""
import TimeTagger
import numpy as np
import matplotlib.pyplot as plt
import time


def gaussian(x, mu, sigma, A):
    """Scaled Gaussian function for visual comparison of the measured data"""
    return A/(np.sqrt(2*np.pi)*sigma)*np.exp(-0.5*(x-mu)**2/sigma**2)

def synchronized_correlation_measurement(tagger, channels, binwidth=1, bins=int(5e5), duration=int(15e12)):
    """
    For the jitter measurement, we use correlation measurements of the periodic built-in test signal.
    The function initializes multiple correlation measurements between channels.
    To have a simultaneous start, we make use of the Synchronized Measurement class.
    Returns correlation histogram data for each channel combination.
    """
    correlation_measurements = {}
    # Here we use a single channel, which is correlated to all other channels.
    start_channel = channels[0]    
    # Use helper class to synchronize measurements
    with TimeTagger.SynchronizedMeasurements(tagger) as sync_meas:
        for stop_channel in channels:
            # When using Synchronizer, correlate only channels within the same device
            if stop_channel - start_channel >= 100:
                start_channel = stop_channel
            
            # Do not correlate the channel against itself
            if start_channel == stop_channel:
                continue
            
            corr_channels =  (start_channel, stop_channel)
            # Initiate the measurements and register them for the synchronized measurement
            corr = TimeTagger.Correlation(sync_meas.getTagger(), start_channel, stop_channel, binwidth, bins)
            correlation_measurements[corr_channels] = corr

        # Start all measurements simultaneously
        sync_meas.startFor(duration)
        sync_meas.waitUntilFinished()

        # Collect correlation results
        corr_histograms = {}
        for corr_channels, corr in correlation_measurements.items():
            corr_histograms[corr_channels] = (corr.getIndex(), corr.getData())
    return corr_histograms

def warmup(tagger, duration=int(30e12)):
    """
    Function to warm-up the Time Tagger until temperatures are stable
    or wait for the specified duration (Time Tagger 20).
    """
    all_channels = tagger.getChannelList(TimeTagger.ChannelEdge.Rising)
    tagger.setTestSignal(all_channels, True)
    # Create load on all channels to warm up the Time Tagger
    countrate = TimeTagger.Countrate(tagger, all_channels)
    if tagger.getModel() == "Time Tagger 20":
        warmup_time = duration/1e12
        print("Warming up ... ", end=' ')
        while warmup_time > 0:
            print(f'{warmup_time:0.0f}', end=' ')
            time.sleep(2)
            warmup_time -= 2
        print("Done.")
    else:
        print("For the Time Tagger Ultra and Time Tagger X, we can check for stable temperatures on the board.")
        warmed_up = False
        cnt = 0
        while not warmed_up:
            pcb_temperatures = []
            fpga_temperatures = []
            # Measure 5 times, then compare if temperature has been stable. If not, repeat.
            for i in range(5):
                time.sleep(1)
                cnt = cnt + 1
                sensor_data = tagger.getSensorData()
                pcb = sensor_data[0]['FPGA board']['Board Temp #1']
                fpga = sensor_data[0]['FPGA board']['FPGA Temp']
                pcb_temperatures.append(pcb)
                fpga_temperatures.append(fpga)
                print("t: {:3d} s, Board temperature: {:2.1f} °C, FPGA temperature: {:2.1f} °C".format(cnt, pcb, fpga))
            if ((max(pcb_temperatures) - min(pcb_temperatures)) < 1
                    and (max(fpga_temperatures) - min(fpga_temperatures)) < 1.5):
                warmed_up = True
                print("The Time Tagger is warmed up.")
                print("")
            else:
                print("The Time Tagger is not yet warmed up.")

def get_jitter_and_delay(hist_t, hist_c):
    """
    Calculate RMS jitter and mean (giving the internal delay of the test signal to later shift the plots)
    """
    mean_t = np.average(hist_t, weights=hist_c)
    jitter_rms = np.sqrt(np.average((hist_t-mean_t)**2, weights=hist_c))
    return jitter_rms, mean_t


# Searching for available Time Taggers
serials = TimeTagger.scanTimeTagger()
print('Found {} connected Time Tagger(s)'.format(len(serials)))

# Connect to Time Tagger and collect device information.
# Please specify serial number explicitly in case more than one device is connected
# and not already taken care of by a synchronizer.
tagger = TimeTagger.createTimeTagger(serial='')
serial = tagger.getSerial() 
model = tagger.getModel()
lic_info = tagger.getDeviceLicense()
edition = lic_info['edition']

if model == 'Time Tagger Ultra' and edition == 'HighRes':
    resolution_modes = ['Standard', 'HighResA', 'HighResB', 'HighResC']
else:
    resolution_modes = ['Standard']

print('Time Tagger info: ')
print(' Serial:          ', serial)
print(' Model:           ', model)
print(' Edition:         ', edition)
print(' Resolution modes:', ', '.join(resolution_modes))

if model not in ['Time Tagger 20', 'Time Tagger Ultra', 'Time Tagger X']:
    raise ValueError('Currently "%s" is not supported by this example script', model)

time.sleep(0.5)
print('Warming up the Time Tagger to get accurate jitter measurement.')
warmup(tagger)

# Storing our data in lists to account for the different modes of the HighRes option.
measured_jitters = {}
within_specs = {}
measured_channels = {}

# We loop over the different HighRes modes, if available. Or only use the single available mode.
for i, mode in enumerate(resolution_modes):
    measured_jitters[mode] = []
    within_specs[mode] = []
    measured_channels[mode] = []
    
    print('Connecting to Time Tagger in {} resolution mode'.format(mode))
    # To change the resolution mode, first we need to disconnect the Time Tagger
    TimeTagger.freeTimeTagger(tagger) 
    resolution = getattr(TimeTagger.Resolution, mode)  # get enum value
    tagger = TimeTagger.createTimeTagger(serial=serial, resolution=resolution)
    
    assert isinstance(tagger,TimeTagger.TimeTagger)
    
    # Determine channel resolutions specification from the device
    tagger_configuration = tagger.getConfiguration()
    specified_jitter_rms = {}
    for input_cfg in tagger_configuration['inputs']:
        input_channel = input_cfg['channel'][0]
        # Retrieve channel rms resolution value (a constant value from device specifications)
        specified_jitter_rms[input_channel] = input_cfg['resolution rms']

    if mode == 'Standard':
        available_inputs = tagger.getChannelList(TimeTagger.ChannelEdge.Rising)
    else:
        # Measure only on HighRes channels
        available_inputs = tagger.getChannelList(TimeTagger.ChannelEdge.HighResRising)

    print(' Available inputs:', available_inputs)

    # Enable test signals on the available inputs
    tagger.setTestSignal(available_inputs, True)  

    print('Measuring for 30 seconds...')
    # Retrieving the measured data
    correlation_data = synchronized_correlation_measurement(tagger, available_inputs, duration=int(30e12))

    print('Measurement complete.\nNow evaluating the data')
    time.sleep(1)

    fig, ax = plt.subplots()  # Create a plot to visualize the TWO channel jitter
    ax.set_title('Visual comparison of the measured two-channel jitters to specifications')
    ax.set_xlabel('Time (ps)')
    ax.set_ylabel('Normalized counts')
    # Looping over the measurement data, evaluating the single channel RMS jitter from it and displaying the results
    # For the visual comparison a Gaussian with standard deviation = sqrt(2)*specified_RMS_jitter is used
    # since we look at the two-channel jitter
    for corr_channels, (hist_t, hist_c) in correlation_data.items():
        start_channel, stop_channel = corr_channels
        rms_t, mean_t = get_jitter_and_delay(hist_t, hist_c)
        jitter_measured = rms_t / np.sqrt(2)  # per channel jitter

        # Plot measure correlation curve
        label = f'{start_channel:>2d} vs {stop_channel:>2d}'
        line, = ax.plot(hist_t-mean_t, hist_c)

        # Check if measured jitter is within specifications
        jitter_spec = specified_jitter_rms[stop_channel]
        jitter_ok = jitter_measured < jitter_spec
        print(f'Correlation {label} Measured channel jitter: {jitter_measured:0.1f} ps    Within specs: {jitter_ok}')

        measured_channels[mode].append(label)
        measured_jitters[mode].append(jitter_measured)
        within_specs[mode].append(jitter_ok)

    line.set_label(label) # Add only one label to prevent crowded plot

    # Plot Gaussian function corresponding to specified jitter
    hist_c_model = gaussian(x=hist_t, mu=0, sigma=jitter_spec*np.sqrt(2), A=np.sum(hist_c))
    ax.plot(hist_t, hist_c_model, color='k', ls='--', label='Specified jitter')
    
    ax.set_xlim(-5*jitter_spec, 5*jitter_spec)
    ax.legend(loc=1)
    plt.show()
    print('\n')

print('Plotting a summary of the results')
# Summary of the measured RMS jitters for a single channel in a table

color_map = {True:'green', False:'red'}  # Cell colors
colLabels = ['Channels', 'single channel\nRMS jitter (ps)', 'within\nspecifications']
for i, mode in enumerate(resolution_modes):
    fig, ax = plt.subplots()
    rowLabels = measured_channels[mode]
    jitters = measured_jitters[mode]
    jitter_ok = within_specs[mode]
    
    cellContent = [(ch, f'{jitt:0.1f}', ok) for ch, jitt, ok in zip(rowLabels, jitters, jitter_ok)]
    cellColors = [('white', 'white', color_map[ok]) for ok in within_specs[mode]]
    the_table = ax.table(
        cellText=cellContent, colLabels=colLabels, cellColours=cellColors,
        loc='center', cellLoc='center'
    )
    the_table.scale(1, 1.5)
    ax.axis('tight')
    ax.axis('off')
    fig.suptitle(f'{model} {edition} {mode}')
    fig.tight_layout()

TimeTagger.freeTimeTagger(tagger)  # Freeing the Time Tagger again so that it can be used in another application
plt.show()

